/*
 * Copyright (c) 2015 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*global define, tizen, console*/

/**
 * Sound profiles model module.
 * Provides functions which allow to manage sound profiles, as well as
 * change current volume level for each system sound type.
 *
 * @module models/soundProfiles
 * @requires {@link core/event}
 * @requires {@link core/storage/localstorage}
 * @namespace models/soundProfiles
 * @memberof models
 */

define({
    name: 'models/soundProfiles',
    requires: [
        'core/event',
        'core/storage/localstorage'
    ],
    def: function soundProfiles(req) {
        'use strict';

        /**
         * Events module instance.
         *
         * @private
         * @type {Module}
         */
        var event = req.core.event,

            /**
             * Local storage module instance.
             *
             * @private
             * @type {Module}
             */
            storage = req.core.storage.localstorage,

            /**
             * Available system sound types list.
             *
             * @private
             * @const {string[]}
             */
            SOUND_TYPES = [
                'SYSTEM',
                'NOTIFICATION',
                'ALARM',
                'MEDIA',
                'RINGTONE'
            ],

            /**
             * Name of storage holding profiles.
             * @const {string}
             */
            STORAGE_KEY = 'soundProfiles',

            /**
             * Contains data and settings for current profile.
             *
             * @private
             * @type {object}
             */
            currentProfile = {},

            /**
             * User's profiles.
             *
             * A profile is an object composed of the following fields:
             *   - id (number) Identifier of the profile.
             *   - title (string) Name of the profile.
             *   - SYSTEM (float) Volume of system sound.
             *   - NOTIFICATION (float) Volume of notification sound.
             *   - ALARM (float) Volume of alarm sound.
             *   - MEDIA (float) Volume of media sound.
             *   - VOICE (float) Volume of voice sound.
             *   - RINGTONE (float) Volume of ringtone sound.
             *
             * @private
             * @type {object[]}
             */
            profiles = [],

            /**
             * ID of active sound profile.
             *
             * @private
             * @type {number}
             */
            activeProfileId = null;

        /**
         * Returns list of sound profiles.
         *
         * @memberof models/soundProfiles
         * @public
         * @returns {object[]}
         */
        function getProfiles() {
            return profiles;
        }

        /**
         * Handles volume change event.
         *
         * @private
         * @param {string} soundType
         * @param {number} volume
         * @fires models.soundProfiles.changed
         */
        function onVolumeChanged(soundType, volume) {
            if (currentProfile[soundType] !== volume) {
                currentProfile[soundType] = volume;
                if (activeProfileId !== null) {
                    activeProfileId = null;
                    event.fire('changed');
                }
            }
        }

        /**
         * Returns list of available sound types.
         *
         * @memberof models/soundProfiles
         * @public
         * @returns {string[]}
         */
        function getSoundTypes() {
            return SOUND_TYPES;
        }

        /**
         * Returns sound profile with specified id.
         * If profile cannot be found, null value is returned.
         *
         * @memberof models/soundProfiles
         * @public
         * @param {number} id Identifier of the profile.
         * @returns {object}
         */
        function getProfile(id) {
            var i = 0,
                profile = null;

            for (i = 0; i < profiles.length; i += 1) {
                profile = profiles[i];
                if (profile.id === id) {
                    return profile;
                }
            }

            return null;
        }

        /**
         * Returns sound profile with a specified title.
         * If profile cannot be found, null value is returned.
         *
         * @memberof models/soundProfiles
         * @public
         * @param {string} title Profile's title.
         * @returns {object}
         */
        function getProfileByTitle(title) {
            var i = 0,
                profile = null;

            for (i = 0; i < profiles.length; i += 1) {
                profile = profiles[i];
                if (profile.title === title) {
                    return profile;
                }
            }

            return null;
        }

        /**
         * Writes profiles to the storage.
         * Returns true if operation completes successfully, false otherwise.
         *
         * @private
         * @returns {boolean}
         */
        function save() {
            return storage.set(STORAGE_KEY, {
                active: activeProfileId,
                profiles: profiles
            });
        }

        /**
         * Changes volume of specified system sound type.
         *
         * @private
         * @param {string} soundType
         * @param {number} volume
         */
        function changeSystemSoundVolume(soundType, volume) {
            try {
                tizen.sound.setVolume(soundType, volume);
                currentProfile[soundType] = tizen.sound.getVolume(soundType);
            } catch (e) {
                console.error(
                    'Setting ' + soundType + ' error: ' + e.message
                );
            }
        }

        /**
         * Updates volume value in a specified sound profile and writes
         * changes to local storage.
         * Returns true if update finished successfully, false otherwise.
         *
         * @memberof models/soundProfiles
         * @public
         * @param {number} profileId Identifier of the profile.
         * @param {string} soundType Type of sound.
         * @param {number} volume Volume value.
         * @returns {boolean}
         */
        function updateProfile(profileId, soundType, volume) {
            var profile = getProfile(profileId);

            if (profile && profile[soundType] !== volume) {
                profile[soundType] = volume;
                if (profile.id === activeProfileId) {
                    changeSystemSoundVolume(soundType, volume);
                }
                return save();
            }
            return false;
        }

        /**
         * Creates new profile based on current system sound settings.
         * Returns true if creation finished successfully, false otherwise.
         *
         * @param {string} [title] Created profile's title. If is not defined
         * the title will be generated.
         * @fires models.soundProfiles.changed
         * @returns {boolean} .
         */
        function createProfile(title) {
            var currentProfileCopy = JSON.parse(JSON.stringify(currentProfile)),
                id = profiles.length ? profiles[profiles.length - 1].id + 1 : 0;

            currentProfileCopy.title = title || 'Profile ' + (id + 1);
            currentProfileCopy.id = id;
            profiles.push(currentProfileCopy);
            if (save()) {
                event.fire('changed');
                return true;
            }
            return false;
        }

        /**
         * Returns index of a profile in the profiles array or -1 if the array
         * doesn't contain the profile with specified id.
         *
         * @private
         * @param {number} profileId Profile's ID.
         * @returns {number}
         */
        function getProfileIndex(profileId) {
            var i = 0;

            for (i = 0; i < profiles.length; i += 1) {
                if (profiles[i].id === profileId) {
                    return i;
                }
            }

            return -1;
        }

        /**
         * Deletes profile.
         *
         * @memberof models/soundProfiles
         * @public
         * @param {number} id Profile's ID.
         * @fires models.soundProfiles.changed
         */
        function deleteProfile(id) {
            var index = getProfileIndex(id);

            if (index !== -1) {
                profiles.splice(index, 1);
            }
            if (save()) {
                event.fire('changed');
            }
        }

        /**
         * Updates current system sound settings with values
         * from specified profile.
         *
         * @private
         * @param {object} profile
         */
        function changeSystemProfile(profile) {
            var i = 0,
                soundType = null;

            for (i = 0; i < SOUND_TYPES.length; i += 1) {
                soundType = SOUND_TYPES[i];
                changeSystemSoundVolume(soundType, profile[soundType]);
            }
        }

        /**
         * Sets active profile.
         *
         * @memberof models/soundProfiles
         * @public
         * @param {number} id Profile's ID.
         * @fires models.soundProfiles.changed
         */
        function setActive(id) {
            if (activeProfileId !== id) {
                changeSystemProfile(getProfile(id));
                activeProfileId = id;
                if (save()) {
                    event.fire('changed');
                }
            }
        }

        /**
         * Returns an object containing stored profiles and id of the active
         * profile or an empty object if there are no saved profiles in the
         * storage.
         *
         * @private
         * @returns {object}
         */
        function getProfilesFromStorage() {
            return storage.get(STORAGE_KEY) || {profiles: [], active: 0};
        }

        /**
         * Initializes module.
         * Loads profiles from local storage, reads current system sound
         * settings and registers volume change listener.
         *
         * @memberof models/soundProfiles
         * @public
         */
        function init() {
            var i = 0,
                soundType = null,
                storedData = getProfilesFromStorage();

            profiles = storedData.profiles;
            activeProfileId = parseInt(storedData.active, 10);

            for (i = 0; i < SOUND_TYPES.length; i += 1) {
                soundType = SOUND_TYPES[i];
                try {
                    currentProfile[soundType] =
                        tizen.sound.getVolume(soundType);
                } catch (e) {
                    console.error(e.message);
                }
            }

            try {
                tizen.sound.setVolumeChangeListener(onVolumeChanged);
            } catch (e) {
                console.error(e.message);
            }
        }

        /**
         * Returns ID of currently active profile.
         *
         * @memberof models/soundProfiles
         * @public
         * @returns {number}
         */
        function getActiveProfileId() {
            return activeProfileId;
        }

        return {
            init: init,
            getSoundTypes: getSoundTypes,
            getProfiles: getProfiles,
            getActiveProfileId: getActiveProfileId,
            getProfile: getProfile,
            getProfileByTitle: getProfileByTitle,
            updateProfile: updateProfile,
            createProfile: createProfile,
            deleteProfile: deleteProfile,
            setActive: setActive
        };
    }
});
